comment %

support file for putting the CPU into flat real mode
support file for gate A20 switching

%



        .DOSSEG
        .MODEL  small, c, os_dos

.386
.CODE

        INCLUDE constant.inc

himemInit     proc public
; return carry set if himem.sys is installed (actually, emm386 is the culprit)
; if it is, we cannot put the CPU into 4G memory mode
; if it is not, force the CPU into flat real mode with 4G segments

        push    esi
	call	himemDetect
	jc	@f

        call    set4gig
        jc      @f

        call    enableA20
        jc      @f

        push    ax
        mov     ax, cs          ; setting 4G screws up ES,DS
   ;     mov     ds, ax          ; restore back to ES=CS=DS
   ;     mov     es, ax
        xor     ax, ax          ; Set segment register to 0
        mov     gs, ax          ; Force GS == 0     
        pop     ax

        clc                     ; exit w/ everything OK
@@:
        pop     esi
        ret
himemInit     endp



himemDetect     proc

	push    ax
        mov     ax, 4300h
        int     2fh
        cmp     al, 80h                     ; al=80h if himem.sys is going
        pop     ax
        stc                             ; assume error
        jz      @f
        clc
@@:
        ret
himemDetect     endp


;----------------------------------------------------------------------------
;                      MASSiVE Flat Real mode routines
;                               (Version 1.0B)
;                                    by
;                               NiX/MASSiVE
;                            Butterfly/MASSiVE
;
; You may use this sourcecode for free, we don't accept anything in return.
; But we would really like it if you greet us, or send your newest productions
; to us.
;
; This is a beta version, only providing you with the essential routines and
; some examples, just real plain flat-real-mode. More special things will be
; available in a next version.
;
; Read the DOC for more information.

.386P    ; 386 Code ofcourse...

;---------------------------------------------------------------------------
; Descriptors and stuff used for setting up FRM
mem32_GDT       dw      4 dup(0)

Code16GDT = 8
                dw      0ffffh,0ffh,9a00h,0
Data16GDT = 16
                dw      0ffffh,0ffh,9200h,0
Data32GDT = 24  
                dw      0ffffh,0ffh,9200h,8fh

GDTptr label fword
                dw      offset GDTptr-1-offset Mem32_GDT
                dd      offset Mem32_GDT ; Absolute adress GDTtable
                dw      0

;-------------------------------------------------------------------------
; This routine sets DS, ES, FS and GS to a 4Gigabyte limit.
; To do this it swaps to PM for a little while, this is not possible in V86
; mode. If computer is in V86 mode the carry will be set on return, and you
; are still in RM.
; Crashes segment registers!!!



Assume ds:@code



Set4Gig:        mov     eax,cr0   ; Check for V86 mode
                ror     eax,1     ;
                jc      Leave4Gig ; If carry set get out of here.

                mov     ax,cs     ; Set up GDT for this code segment
                mov     ds,ax
                movzx   eax,ax
                shl     eax,4
                add     dword ptr ds:[GDTptr+2],eax
                lgdt    fword ptr ds:[GDTptr]

                mov     ax,cs
                and     eax,65535
                shl     eax,4
                mov     word ptr ds:Mem32_GDT[Code16GDT+2],ax
                mov     word ptr ds:Mem32_GDT[Data16GDT+2],ax
                ror     eax,16
                mov     byte ptr ds:Mem32_GDT[Code16GDT+4],al
                mov     byte ptr ds:Mem32_GDT[Data16GDT+4],al
                mov     byte ptr ds:Mem32_GDT[Code16GDT+7],ah
                mov     byte ptr ds:Mem32_GDT[Data16GDT+7],ah

                cli    				; No ints

; Beam us to PMODE!!
                mov     eax,cr0
                or      al,1
                mov     cr0,eax

                db      0eah               	; Far jump to PMODE label
                dw      offset Pmode
                dw      Code16GDT

; We're in PMODE!!
Pmode:          mov     ax,Data32GDT
                mov     ds,ax   	   	; Set them all to 4Gig
                mov     es,ax   
                mov     fs,ax   
                mov     gs,ax
                
; Beaming out!
                mov     eax,cr0
                and     al,0feh
                mov     cr0,eax

                db      0eah           		; Far jump to Rmode label.
                dw      offset Rmode
                dw      @code

; Back in the real world!
Rmode:          clc  ; No carry, everything went OK
                sti  ; Ints Back on.
                
Leave4Gig:      ret






;-----------------------------------------------------------------------------
; Enables the A20 adress line. This is needed to adress the memory above 1Mb.
; Do this at the start of your program and everytime another program has been
; active. In an interrupt routine this means you should do it everytime the
; routine is being called, don't worry, it doesn't takes much time.
EnableA20:
                push    ax
                mov     al,0D1h
                out     64h, al
                call    A20wait
                mov     al,0DFh
                out     60h, al
                call    A20wait
                mov     al,0FFh
                out     64h,al
                call    A20wait
                clc
                pop     ax
                ret

A20Wait:        in      al,64h
                jmp     $+2
                and     al,2
                jnz     A20wait
                ret



End
